// Copyright 2021 Sony Group Corporation.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#ifndef NBLA_FUNCTION_BOOL_SCATTER_HPP
#define NBLA_FUNCTION_BOOL_SCATTER_HPP

#include <nbla/cpu.hpp>
#include <nbla/function.hpp>
#include <nbla/function_registry.hpp>

namespace nbla {

NBLA_REGISTER_FUNCTION_HEADER(BoolScatter);

/**
    @todo Write doc.

Inputs:

Outputs:

\ingroup FunctionImplGrp
 */
template <typename T> class BoolScatter : public BaseFunction<> {

public:
  BoolScatter(const Context &ctx) : BaseFunction(ctx) {}

  virtual ~BoolScatter() {}
  virtual shared_ptr<Function> copy() const { return create_BoolScatter(ctx_); }
  virtual int min_inputs() { return 2; }
  virtual int min_outputs() { return 1; }
  virtual vector<dtypes> in_types() {
    return vector<dtypes>{get_dtype<T>(), get_dtype<T>(), get_dtype<T>()};
  }
  virtual vector<dtypes> out_types() { return vector<dtypes>{get_dtype<T>()}; }
  virtual vector<string> allowed_array_classes() {
    return SingletonManager::get<Cpu>()->array_classes();
  }
  virtual string name() { return "BoolScatter"; }

  virtual int inplace_data(int i) const {
    return i == 2 ? Function::INPLACE : Function::NOT_INPLACE;
  }
  virtual int inplace_data_with(int i) const { return 0; }

protected:
  NBLA_API virtual void setup_impl(const Variables &inputs,
                                   const Variables &outputs);
  NBLA_API virtual void forward_impl(const Variables &inputs,
                                     const Variables &outputs);
  NBLA_API virtual void backward_impl(const Variables &inputs,
                                      const Variables &outputs,
                                      const vector<bool> &propagate_down,
                                      const vector<bool> &accum);
  virtual bool grad_depends_input_data_impl(int i, int j) const {
    if (i == 0 && j == 1)
      return true;
    if (i == 2 && j == 1)
      return true;
    return false;
  }
  virtual bool auto_grad_depends_input_data_impl(int i, int j) const {
    // bool_scatter_backward requires inputs[1].
    if (j == 1) {
      return true;
    }
    return false;
  }

  virtual bool grad_depends_output_data(int i, int o) const { return false; }
};
}
#endif
